# -*- coding: utf-8 -*-
"""
Core REST API code.

This package exists so we can separate the REST API MVP from the rest of the
code. The code that lives in this package should not require anything else to
run and should also provide a minimal set of tools for building an API.

All advanced functionality should be implemented in other packages.
"""
from __future__ import absolute_import, unicode_literals



# -*- coding: utf-8 -*-
"""

Integration with webapp2.

All webapp2 dependant code should live in this package. webapp2 is the default
framework used by Google AppEngine so it makes restible based apps easy to
implement for this platform.
"""
from __future__ import absolute_import, unicode_literals
from .endpoint import Webapp2Endpoint



# -*- coding: utf-8 -*-
"""
Base class for resources using Google AppEngine ndb as storage.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from logging import getLogger

# 3rd party imports
from six import iteritems
from serafin import Fieldspec
from sqlalchemy.exc import SQLAlchemyError, IntegrityError

# local imports
from restible.core.model import ModelResource


L = getLogger(__name__)


class SqlAlchemyResource(ModelResource):
    """ Base class for ndb based resources.

    This provides a basic implementation that can be used out of the box. It
    provides no authentication/authorization.
    """
    name = None
    model = None
    spec = Fieldspec('*')
    schema = {}
    read_only = []
    _public_props = None
    _db_session = None

    @classmethod
    def init_session(cls, db_session):
        """ Initialize SQLAlchemy resources. """
        cls._db_session = db_session

    @property
    def db_session(self):
        """ Property returning the current SQLAlchemy session.

        Will try to get it from ``self.model.query.session`` as this is where
        flask has it.
        """
        if hasattr(self.model, 'query'):
            return self.model.query.session
        else:
            return self._db_session

    def create_item(self, values):
        """ Create new model item. """
        try:
            item = self.model(**values)

            self.db_session.add(item)
            self.db_session.commit()

            return item
        except IntegrityError:
            self.db_session.rollback()
            raise ModelResource.AlreadyExists()
        except SQLAlchemyError:
            self.db_session.rollback()
            raise

    def update_item(self, request, values):
        item = self.get_requested(request)

        if item is None:
            return None

        for name, value in iteritems(values):
            if name not in ('id',):
                try:
                    setattr(item, name, value)
                except AttributeError:
                    L.exception("Failed to set attribute '{}'".format(name))
                    raise

        try:
            self.db_session.commit()
        except SQLAlchemyError:
            self.db_session.rollback()
            raise

        return item

    def delete_item(self, item):
        try:
            self.db_session.delete(item)
            self.db_session.commit()
        except SQLAlchemyError:
            self.db_session.rollback()
            raise

    def deserialize(self, data):
        """ Convert JSON data into model field types.

        The value returned by this function can be used directly to create new
        item and update existing ones.
        """
        return {n: self.get_field_value(n, v) for n, v in iteritems(data)}

    def get_field_value(self, name, value):
        """ Coerce value to a model field compatible representation. """
        return value

    def dbquery(self, request, filters):
        """ Return a model query with the given filters.

        The query can be further customised like any ndb query.

        :return google.appengine.ext.ndb.Query:
            The query with the given filters already applied.
        """
        filters = [getattr(self.model, n) == v for n, v in iteritems(filters)]
        return self.get_queryset(request).filter(*filters).all()

    def get_requested(self, request):
        """ Get requested item. """
        pk = self.get_pk(request)
        return self.get_queryset(request).filter(
            self.model.id == pk
        ).one_or_none()

    def get_queryset(self, request):
        """ Extension point for one place to limit the data returned.

        Both .dbquery() and .get_requested() use this function as the base
        queryset. This means that if you need to restrict access to some db
        entries, you can just overload this method.

        :param request:
            The request associated with the call. Not used by default, but can
            be handy when overriding.
        :return:
            The SQLAlchemy resource.
        """
        return self.model.query



# -*- coding: utf-8 -*-
"""
REST resource base class.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import inspect
import re

from .routing import api_action


class RestResource(object):
    """ Represents the operations available on the given resource.

    This class should not handle HTTP directly. Rather than that there should
    be a separate implementation that maps HTTP layer onto resources. This way
    we can swap the HTTP layer without reimplementing the resource. This also
    follows SRP.

    If the given `RestResource` subclass doesn't have one of the REST methods
    implemented, the endpoint should return 405 Method not allowed (or 404?)
    """

    _NAME_RE = re.compile(r'^[a-zA-Z_][\w\d_]*$')
    name = None
    parent = None

    def __init__(self):
        self._validate_name(self.name)

    def rest_actions(self):
        """ Return all actions defined on the current resource.

        :return List[function]:
            List of action handlers. To get the metadata for a given *action*
            just use ``api_action.get_meta(action)``
        """
        ATTR = '_actions'

        if not hasattr(self, ATTR):
            actions = []
            for _, method in inspect.getmembers(self, inspect.ismethod):
                if api_action.is_action(method):
                    actions.append(method)

            setattr(self, ATTR, actions)

        return getattr(self, ATTR)

    def get_pk(self, request):
        """ Read the current resource type PK from the request.

        :param HttpRequest request:
            An HTTP request. Must contain the ``.rest_keys`` dictionary injected
            by `RestEndpoint.dispatch()` method.
        :return:
            Primary key that is associated with this resource type. This is
            extracted from ``.rest_keys`` based on the resource name.
        """
        if hasattr(request, 'rest_keys'):
            return request.rest_keys.get(self.name + '_pk')

    def _validate_name(self, name):
        """ Validate the name is a valid resource name. """
        if name is None or not self._NAME_RE.match(name):
            raise ValueError(
                'REST endpoint name contain only letters, digits and '
                'underscores and cannot start with a digit'
            )

    def get_requested(self, request):
        """ Get an item associated with the request.

        This is used by all detail views/actions to get the item that the
        request is concerned with (usually from the URL). This is an
        implementation detail and is highly dependant on the underlying web
        framework used.

        :param request:
            HTTP request.
        :return RestResource:
            The item associated with the request.
        """
        raise NotImplementedError("{}.get_requested() not implemented".format(
            self.__class__.__name__
        ))

    def rest_query(self, request, params):
        """ GET list. """
        raise NotImplementedError(".rest_query() not implemented".format(
            self.__class__.__name__
        ))

    def rest_get(self, request, params):
        """ GET detail. """
        raise NotImplementedError("{}.rest_get() not implemented".format(
            self.__class__.__name__
        ))

    def rest_create(self, request, data):
        """ POST list. """
        raise NotImplementedError("{}.rest_create() not implemented".format(
            self.__class__.__name__
        ))

    def rest_update(self, request, data):
        """ PUT detail. """
        raise NotImplementedError("{}.rest_update() not implemented".format(
            self.__class__.__name__
        ))

    def rest_delete(self, request):
        """ DELETE detail. """
        raise NotImplementedError("{}.rest_delete() not implemented".format(
            self.__class__.__name__
        ))

    def rest_options(self, request):
        """ OPTIONS list/detail. """
        raise NotImplementedError("{}.rest_options() not implemented".format(
            self.__class__.__name__
        ))

    def rest_head(self, request):
        """ OPTIONS list/detail. """
        raise NotImplementedError("{}.rest_head() not implemented".format(
            self.__class__.__name__
        ))



# -*- coding: utf-8 -*-
"""
Helpers for setting up mapping between URLs and resources.
"""
from __future__ import absolute_import, unicode_literals
from jsobj import jsobj


class api_route(object):
    """ Decorator for defining API routs.

    API routes are endpoints that do not belong to any resource.
    """
    ROUTE_ATTR = '__api_route__'

    def __init__(self, methods=None):
        self.meta = jsobj({
            'methods': methods or ['post']
        })

    def __call__(self, fn):
        """ Decorator. """
        setattr(fn, self.ROUTE_ATTR, self.meta)
        return fn

    @classmethod
    def get_meta(cls, route):
        """ Get route metadata. """
        return getattr(route, cls.ROUTE_ATTR, {})


class api_action(object):
    """ Decorator for defining resource actions.

    Actions are extra HTTP endpoints that can be called with POST.
    """
    ACTION_ATTR = '__api_action__'

    def __init__(self, name=None, generic=False, protected=True, methods=None):
        self.name = name
        self.generic = generic
        self.protected = protected
        self.methods = [m.lower() for m in (methods or ['post'])]

    def __call__(self, fn):
        """ Decorator. """
        setattr(fn, self.ACTION_ATTR, jsobj({
            'name': self.name or fn.__name__,
            'generic': self.generic,
            'protected': self.protected,
            'methods': self.methods,
        }))
        return fn

    @classmethod
    def is_action(cls, obj):
        """ Return true if obj is an API action. """
        return hasattr(obj, cls.ACTION_ATTR)

    @classmethod
    def get_meta(cls, action):
        """ Get action metadata """
        return getattr(action, cls.ACTION_ATTR, {})



# -*- coding: utf-8 -*-
"""
Django implementation of a `RestEndpoint`.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json

# 3rd party imports
from django.conf.urls import url
from django.http import JsonResponse
from six import iteritems

# local imports
from restible import api_action, RestEndpoint, RestResource, RawResponse


class DjangoEndpoint(RestEndpoint):
    """ Endpoint implementation to use in django projects. """

    def dispatch(self, request, **route_params):
        """ Django 'calls' views so this is what will be ran by django.

        This should cover the whole request lifecycle.
        """
        request.rest_keys = route_params

        result = self.call_rest_handler(request.method, request)
        return self.response_from_result(result)

    def dispatch_action(self, request, name, generic, **route_params):
        """ Dispatch resource action. """
        request.rest_keys = route_params

        result = self.call_action_handler(request, name, generic)
        return self.response_from_result(result)

    def response_from_result(self, result):
        """ Create django response from RestResult.

        :param RestResult result:
        :return JsonResponse:
            Django response instance, ready to return from the view.
        """
        if isinstance(result, RawResponse):
            return result.response

        response = JsonResponse(
            status=result.status,
            safe=False,
            data=result.data
        )
        for key, value in iteritems(result.headers):
            response[key] = value

        return response

    @classmethod
    def extract_request_data(cls, request):
        try:
            if request.body:
                body = request.body.decode('utf-8')
                return json.loads(body)
            else:
                return None
        except:
            raise ValueError("Invalid JSON body: '{}'".format(
                request.body.decode('utf-8')
            ))

    @classmethod
    def extract_request_query_string(cls, request):
        return request.GET

    def urls(self, base_url):
        """ Return all the URLs required by the endpoint.

        :return list<url>:
            A list of URLs that can be directly appended to ``urlpatterns`` of
            choice.
        """
        # TODO: Using resource.name for both prefix and PK name might blow up.
        # There might be one generic endpoint called /bars and we might want to
        # create a nested one /foo/bars. In those cases we have two resources
        # with the same name - bars.
        #
        # Another question is whether this is an issue. In the above example
        # the resources are not at the same 'level' in the tree. This will
        # definitely fail if we have a recursive endpoint, but that might be
        # even worse.
        if base_url.endswith('/'):
            base_url = base_url[:-1]

        if base_url.startswith('/'):
            base_url = base_url[1:]

        name = self.resource.name
        list_url = base_url
        item_url = '{url}/(?P<{name}_pk>[^/]+)'.format(url=base_url, name=name)
        routes = []

        # actions
        for action in self.resource.rest_actions():
            meta = api_action.get_meta(action)
            route_name = '{}-{}'.format(self.resource.name, meta.name),
            action_url = '{url}/{action}'.format(
                url=list_url if meta.generic else item_url,
                action=meta.name
            )

            routes.append(
                url(action_url, self.dispatch_action, name=route_name, kwargs={
                    'name': meta.name,
                    'generic': meta.generic,
                })
            )

        routes += [
            url(item_url, self.dispatch, name='{name}-item'.format(name=name)),
            url(list_url, self.dispatch, name='{name}-list'.format(name=name)),
        ]

        return routes

    @classmethod
    def make_urls(cls, routes):
        """ Create a list of URLs handled by this endpoint. """
        urlconf = []

        for entry in routes:
            endpoint = None
            opts = {}

            if len(entry) == 2:
                url, res_cls = entry
            else:
                url, res_cls, opts = entry

            if isinstance(res_cls, type) and issubclass(res_cls, RestResource):
                endpoint = cls(res_cls, **opts)
            elif isinstance(res_cls, RestEndpoint):
                endpoint = res_cls
            else:
                raise RuntimeError("Invalid resource class for {}".format(
                    res_cls.name
                ))

            urlconf += endpoint.urls(url)

        return urlconf



# -*- coding: utf-8 -*-
""" Various helpers. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import inspect


def iter_public_props(obj, predicate=None):
    """ Iterate over public properties of an object.

    :param Any obj:
        The object we want to get the properties of.
    :param function predicate:
        Additional predicate to filter out properties we're interested in. This
        function will be called on every property of the object with the
        property name and value as it arguments. If it returns True, the
        property will be yielded by this generator.
    """
    predicate = predicate or (lambda n, v: True)
    obj_type = type(obj)

    if inspect.isclass(obj):
        # This is a class
        for name, value in obj.__dict__.items():
            if isinstance(value, property):
                yield name, value
    else:
        # This is an instance
        for name in dir(obj):
            if name.startswith('_'):
                continue

            member = getattr(obj_type, name)
            if not isinstance(member, property):
                continue

            try:
                value = getattr(obj, name)
                if predicate(name, value):
                    yield name, value
            except AttributeError:
                pass
